@@ -44,9 +44,17 @@
 #include "wayland-private.h"
 #include "wayland-os.h"
 
-#define DIV_ROUNDUP(n, a) ( ((n) + ((a) - 1)) / (a) )
+static inline uint32_t
+div_roundup(uint32_t n, size_t a)
+{
+	/* The cast to uint64_t is necessary to prevent overflow when rounding
+	 * values close to UINT32_MAX. After the division it is again safe to
+	 * cast back to uint32_t.
+	 */
+	return (uint32_t) (((uint64_t) n + (a - 1)) / a);
+}
 
-struct wl_buffer {
+struct wl_ring_buffer {
 	char data[4096];
 	uint32_t head, tail;
 };
@@ -57,14 +65,14 @@ struct wl_buffer {
 #define CLEN		(CMSG_LEN(MAX_FDS_OUT * sizeof(int32_t)))
 
 struct wl_connection {
-	struct wl_buffer in, out;
-	struct wl_buffer fds_in, fds_out;
+	struct wl_ring_buffer in, out;
+	struct wl_ring_buffer fds_in, fds_out;
 	int fd;
 	int want_flush;
 };
 
 static int
-wl_buffer_put(struct wl_buffer *b, const void *data, size_t count)
+ring_buffer_put(struct wl_ring_buffer *b, const void *data, size_t count)
 {
 	uint32_t head, size;
 
@@ -90,7 +98,7 @@ wl_buffer_put(struct wl_buffer *b, const void *data, size_t count)
 }
 
 static void
-wl_buffer_put_iov(struct wl_buffer *b, struct iovec *iov, int *count)
+ring_buffer_put_iov(struct wl_ring_buffer *b, struct iovec *iov, int *count)
 {
 	uint32_t head, tail;
 
@@ -114,7 +122,7 @@ wl_buffer_put_iov(struct wl_buffer *b, struct iovec *iov, int *count)
 }
 
 static void
-wl_buffer_get_iov(struct wl_buffer *b, struct iovec *iov, int *count)
+ring_buffer_get_iov(struct wl_ring_buffer *b, struct iovec *iov, int *count)
 {
 	uint32_t head, tail;
 
@@ -138,7 +146,7 @@ wl_buffer_get_iov(struct wl_buffer *b, struct iovec *iov, int *count)
 }
 
 static void
-wl_buffer_copy(struct wl_buffer *b, void *data, size_t count)
+ring_buffer_copy(struct wl_ring_buffer *b, void *data, size_t count)
 {
 	uint32_t tail, size;
 
@@ -153,7 +161,7 @@ wl_buffer_copy(struct wl_buffer *b, void *data, size_t count)
 }
 
 static uint32_t
-wl_buffer_size(struct wl_buffer *b)
+ring_buffer_size(struct wl_ring_buffer *b)
 {
 	return b->head - b->tail;
 }
@@ -173,24 +181,31 @@ wl_connection_create(int fd)
 }
 
 static void
-close_fds(struct wl_buffer *buffer, int max)
+close_fds(struct wl_ring_buffer *buffer, int max)
 {
 	int32_t fds[sizeof(buffer->data) / sizeof(int32_t)], i, count;
 	size_t size;
 
-	size = buffer->head - buffer->tail;
+	size = ring_buffer_size(buffer);
 	if (size == 0)
 		return;
 
-	wl_buffer_copy(buffer, fds, size);
+	ring_buffer_copy(buffer, fds, size);
 	count = size / sizeof fds[0];
 	if (max > 0 && max < count)
 		count = max;
+	size = count * sizeof fds[0];
 	for (i = 0; i < count; i++)
 		close(fds[i]);
 	buffer->tail += size;
 }
 
+void
+wl_connection_close_fds_in(struct wl_connection *connection, int max)
+{
+	close_fds(&connection->fds_in, max);
+}
+
 int
 wl_connection_destroy(struct wl_connection *connection)
 {
@@ -206,7 +221,7 @@ wl_connection_destroy(struct wl_connection *connection)
 void
 wl_connection_copy(struct wl_connection *connection, void *data, size_t size)
 {
-	wl_buffer_copy(&connection->in, data, size);
+	ring_buffer_copy(&connection->in, data, size);
 }
 
 void
@@ -216,12 +231,12 @@ wl_connection_consume(struct wl_connection *connection, size_t size)
 }
 
 static void
-build_cmsg(struct wl_buffer *buffer, char *data, int *clen)
+build_cmsg(struct wl_ring_buffer *buffer, char *data, int *clen)
 {
 	struct cmsghdr *cmsg;
 	size_t size;
 
-	size = buffer->head - buffer->tail;
+	size = ring_buffer_size(buffer);
 	if (size > MAX_FDS_OUT * sizeof(int32_t))
 		size = MAX_FDS_OUT * sizeof(int32_t);
 
@@ -230,7 +245,7 @@ build_cmsg(struct wl_buffer *buffer, char *data, int *clen)
 		cmsg->cmsg_level = SOL_SOCKET;
 		cmsg->cmsg_type = SCM_RIGHTS;
 		cmsg->cmsg_len = CMSG_LEN(size);
-		wl_buffer_copy(buffer, CMSG_DATA(cmsg), size);
+		ring_buffer_copy(buffer, CMSG_DATA(cmsg), size);
 		*clen = cmsg->cmsg_len;
 	} else {
 		*clen = 0;
@@ -238,7 +253,7 @@ build_cmsg(struct wl_buffer *buffer, char *data, int *clen)
 }
 
 static int
-decode_cmsg(struct wl_buffer *buffer, struct msghdr *msg)
+decode_cmsg(struct wl_ring_buffer *buffer, struct msghdr *msg)
 {
 	struct cmsghdr *cmsg;
 	size_t size, max, i;
@@ -251,13 +266,13 @@ decode_cmsg(struct wl_buffer *buffer, struct msghdr *msg)
 			continue;
 
 		size = cmsg->cmsg_len - CMSG_LEN(0);
-		max = sizeof(buffer->data) - wl_buffer_size(buffer);
+		max = sizeof(buffer->data) - ring_buffer_size(buffer);
 		if (size > max || overflow) {
 			overflow = 1;
 			size /= sizeof(int32_t);
 			for (i = 0; i < size; i++)
 				close(((int*)CMSG_DATA(cmsg))[i]);
-		} else if (wl_buffer_put(buffer, CMSG_DATA(cmsg), size) < 0) {
+		} else if (ring_buffer_put(buffer, CMSG_DATA(cmsg), size) < 0) {
 				return -1;
 		}
 	}
@@ -284,7 +299,7 @@ wl_connection_flush(struct wl_connection *connection)
 
 	tail = connection->out.tail;
 	while (connection->out.head - connection->out.tail > 0) {
-		wl_buffer_get_iov(&connection->out, iov, &count);
+		ring_buffer_get_iov(&connection->out, iov, &count);
 
 		build_cmsg(&connection->fds_out, cmsg, &clen);
 
@@ -317,7 +332,7 @@ wl_connection_flush(struct wl_connection *connection)
 uint32_t
 wl_connection_pending_input(struct wl_connection *connection)
 {
-	return wl_buffer_size(&connection->in);
+	return ring_buffer_size(&connection->in);
 }
 
 int
@@ -328,12 +343,12 @@ wl_connection_read(struct wl_connection *connection)
 	char cmsg[CLEN];
 	int len, count, ret;
 
-	if (wl_buffer_size(&connection->in) >= sizeof(connection->in.data)) {
+	if (ring_buffer_size(&connection->in) >= sizeof(connection->in.data)) {
 		errno = EOVERFLOW;
 		return -1;
 	}
 
-	wl_buffer_put_iov(&connection->in, iov, &count);
+	ring_buffer_put_iov(&connection->in, iov, &count);
 
 	msg.msg_name = NULL;
 	msg.msg_namelen = 0;
@@ -370,7 +385,7 @@ wl_connection_write(struct wl_connection *connection,
 			return -1;
 	}
 
-	if (wl_buffer_put(&connection->out, data, count) < 0)
+	if (ring_buffer_put(&connection->out, data, count) < 0)
 		return -1;
 
 	connection->want_flush = 1;
@@ -389,10 +404,10 @@ wl_connection_queue(struct wl_connection *connection,
 			return -1;
 	}
 
-	return wl_buffer_put(&connection->out, data, count);
+	return ring_buffer_put(&connection->out, data, count);
 }
 
-static int
+int
 wl_message_count_arrays(const struct wl_message *message)
 {
 	int i, arrays;
@@ -414,13 +429,13 @@ wl_connection_get_fd(struct wl_connection *connection)
 static int
 wl_connection_put_fd(struct wl_connection *connection, int32_t fd)
 {
-	if (wl_buffer_size(&connection->fds_out) == MAX_FDS_OUT * sizeof fd) {
+	if (ring_buffer_size(&connection->fds_out) == MAX_FDS_OUT * sizeof fd) {
 		connection->want_flush = 1;
 		if (wl_connection_flush(connection) < 0)
 			return -1;
 	}
 
-	return wl_buffer_put(&connection->fds_out, &fd, sizeof fd);
+	return ring_buffer_put(&connection->fds_out, &fd, sizeof fd);
 }
 
 const char *
@@ -523,6 +538,64 @@ wl_argument_from_va_list(const char *signature, union wl_argument *args,
 	}
 }
 
+static void
+wl_closure_clear_fds(struct wl_closure *closure)
+{
+	const char *signature = closure->message->signature;
+	struct argument_details arg;
+	int i;
+
+	for (i = 0; i < closure->count; i++) {
+		signature = get_next_argument(signature, &arg);
+		if (arg.type == 'h')
+			closure->args[i].h = -1;
+	}
+}
+
+static struct wl_closure *
+wl_closure_init(const struct wl_message *message, uint32_t size,
+                int *num_arrays, union wl_argument *args)
+{
+	struct wl_closure *closure;
+	int count;
+
+	count = arg_count_for_signature(message->signature);
+	if (count > WL_CLOSURE_MAX_ARGS) {
+		wl_log("too many args (%d)\n", count);
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (size) {
+		*num_arrays = wl_message_count_arrays(message);
+		closure = malloc(sizeof *closure + size +
+				 *num_arrays * sizeof(struct wl_array));
+	} else {
+		closure = malloc(sizeof *closure);
+	}
+
+	if (!closure) {
+		errno = ENOMEM;
+		return NULL;
+	}
+
+	if (args)
+		memcpy(closure->args, args, count * sizeof *args);
+
+	closure->message = message;
+	closure->count = count;
+
+	/* Set these all to -1 so we can close any that have been
+	 * set to a real value during wl_closure_destroy().
+	 * We may have copied a bunch of fds into the closure with
+	 * memcpy previously, but those are undup()d client fds
+	 * that we would have replaced anyway.
+	 */
+	wl_closure_clear_fds(closure);
+
+	return closure;
+}
+
 struct wl_closure *
 wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 		   union wl_argument *args,
@@ -534,20 +607,11 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 	const char *signature;
 	struct argument_details arg;
 
-	count = arg_count_for_signature(message->signature);
-	if (count > WL_CLOSURE_MAX_ARGS) {
-		wl_log("too many args (%d)\n", count);
-		errno = EINVAL;
+	closure = wl_closure_init(message, 0, NULL, args);
+	if (closure == NULL)
 		return NULL;
-	}
 
-	closure = malloc(sizeof *closure);
-	if (closure == NULL) {
-		errno = ENOMEM;
-		return NULL;
-	}
-
-	memcpy(closure->args, args, count * sizeof *args);
+	count = closure->count;
 
 	signature = message->signature;
 	for (i = 0; i < count; i++) {
@@ -580,8 +644,12 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 		case 'h':
 			fd = args[i].h;
 			dup_fd = wl_os_dupfd_cloexec(fd, 0);
-			if (dup_fd < 0)
-				wl_abort("dup failed: %s\n", strerror(errno));
+			if (dup_fd < 0) {
+				wl_closure_destroy(closure);
+				wl_log("error marshalling arguments for %s: dup failed: %s\n",
+				       message->name, strerror(errno));
+				return NULL;
+			}
 			closure->args[i].h = dup_fd;
 			break;
 		default:
@@ -592,8 +660,6 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 
 	closure->sender_id = sender->id;
 	closure->opcode = opcode;
-	closure->message = message;
-	closure->count = count;
 
 	return closure;
 
@@ -624,30 +690,30 @@ wl_connection_demarshal(struct wl_connection *connection,
 			struct wl_map *objects,
 			const struct wl_message *message)
 {
-	uint32_t *p, *next, *end, length, id;
+	uint32_t *p, *next, *end, length, length_in_u32, id;
 	int fd;
 	char *s;
-	unsigned int i, count, num_arrays;
+	int i, count, num_arrays;
 	const char *signature;
 	struct argument_details arg;
 	struct wl_closure *closure;
-	struct wl_array *array, *array_extra;
+	struct wl_array *array_extra;
 
-	count = arg_count_for_signature(message->signature);
-	if (count > WL_CLOSURE_MAX_ARGS) {
-		wl_log("too many args (%d)\n", count);
+	/* Space for sender_id and opcode */
+	if (size < 2 * sizeof *p) {
+		wl_log("message too short, invalid header\n");
+		wl_connection_consume(connection, size);
 		errno = EINVAL;
+		return NULL;
+	}
+
+	closure = wl_closure_init(message, size, &num_arrays, NULL);
+	if (closure == NULL) {
 		wl_connection_consume(connection, size);
 		return NULL;
 	}
 
-	num_arrays = wl_message_count_arrays(message);
-	closure = malloc(sizeof *closure + size + num_arrays * sizeof *array);
-	if (closure == NULL) {
-		errno = ENOMEM;
-		wl_connection_consume(connection, size);
-		return NULL;
-	}
+	count = closure->count;
 
 	array_extra = closure->extra;
 	p = (uint32_t *)(closure->extra + num_arrays);
@@ -664,7 +730,8 @@ wl_connection_demarshal(struct wl_connection *connection,
 		if (arg.type != 'h' && p + 1 > end) {
 			wl_log("message too short, "
 			       "object (%d), message %s(%s)\n",
-			       *p, message->name, message->signature);
+			       closure->sender_id, message->name,
+			       message->signature);
 			errno = EINVAL;
 			goto err;
 		}
@@ -682,13 +749,20 @@ wl_connection_demarshal(struct wl_connection *connection,
 		case 's':
 			length = *p++;
 
+			if (length == 0 && !arg.nullable) {
+				wl_log("NULL string received on non-nullable "
+				       "type, message %s(%s)\n", message->name,
+				       message->signature);
+				errno = EINVAL;
+				goto err;
+			}
 			if (length == 0) {
 				closure->args[i].s = NULL;
 				break;
 			}
 
-			next = p + DIV_ROUNDUP(length, sizeof *p);
-			if (next > end) {
+			length_in_u32 = div_roundup(length, sizeof *p);
+			if ((uint32_t) (end - p) < length_in_u32) {
 				wl_log("message too short, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
@@ -696,6 +770,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 				errno = EINVAL;
 				goto err;
 			}
+			next = p + length_in_u32;
 
 			s = (char *) p;
 
@@ -746,8 +821,8 @@ wl_connection_demarshal(struct wl_connection *connection,
 		case 'a':
 			length = *p++;
 
-			next = p + DIV_ROUNDUP(length, sizeof *p);
-			if (next > end) {
+			length_in_u32 = div_roundup(length, sizeof *p);
+			if ((uint32_t) (end - p) < length_in_u32) {
 				wl_log("message too short, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
@@ -755,6 +830,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 				errno = EINVAL;
 				goto err;
 			}
+			next = p + length_in_u32;
 
 			array_extra->size = length;
 			array_extra->alloc = 0;
@@ -773,7 +849,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 				goto err;
 			}
 
-			wl_buffer_copy(&connection->fds_in, &fd, sizeof fd);
+			ring_buffer_copy(&connection->fds_in, &fd, sizeof fd);
 			connection->fds_in.tail += sizeof fd;
 			closure->args[i].h = fd;
 			break;
@@ -783,9 +859,6 @@ wl_connection_demarshal(struct wl_connection *connection,
 		}
 	}
 
-	closure->count = count;
-	closure->message = message;
-
 	wl_connection_consume(connection, size);
 
 	return closure;
@@ -797,6 +870,23 @@ wl_connection_demarshal(struct wl_connection *connection,
 	return NULL;
 }
 
+bool
+wl_object_is_zombie(struct wl_map *map, uint32_t id)
+{
+	uint32_t flags;
+
+	/* Zombie objects only exist on the client side. */
+	if (map->side == WL_MAP_SERVER_SIDE)
+		return false;
+
+	/* Zombie objects can only have been created by the client. */
+	if (id >= WL_SERVER_ID_START)
+		return false;
+
+	flags = wl_map_lookup_flags(map, id);
+	return !!(flags & WL_MAP_ENTRY_ZOMBIE);
+}
+
 int
 wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 {
@@ -818,7 +908,7 @@ wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 			closure->args[i].o = NULL;
 
 			object = wl_map_lookup(objects, id);
-			if (object == WL_ZOMBIE_OBJECT) {
+			if (wl_object_is_zombie(objects, id)) {
 				/* references object we've already
 				 * destroyed client side */
 				object = NULL;
@@ -933,6 +1023,8 @@ wl_closure_invoke(struct wl_closure *closure, uint32_t flags,
 			 opcode, target->interface->name);
 	}
 	ffi_call(&cif, implementation[opcode], NULL, ffi_args);
+
+	wl_closure_clear_fds(closure);
 }
 
 void
@@ -941,6 +1033,8 @@ wl_closure_dispatch(struct wl_closure *closure, wl_dispatcher_func_t dispatcher,
 {
 	dispatcher(target->implementation, target, opcode, closure->message,
 		   closure->args);
+
+	wl_closure_clear_fds(closure);
 }
 
 static int
@@ -962,9 +1056,10 @@ copy_fds_to_connection(struct wl_closure *closure,
 		fd = closure->args[i].h;
 		if (wl_connection_put_fd(connection, fd)) {
 			wl_log("request could not be marshaled: "
-			       "can't send file descriptor");
+			       "can't send file descriptor\n");
 			return -1;
 		}
+		closure->args[i].h = -1;
 	}
 
 	return 0;
@@ -1002,7 +1097,7 @@ buffer_size_for_closure(struct wl_closure *closure)
 			}
 
 			size = strlen(closure->args[i].s) + 1;
-			buffer_size += 1 + DIV_ROUNDUP(size, sizeof(uint32_t));
+			buffer_size += 1 + div_roundup(size, sizeof(uint32_t));
 			break;
 		case 'a':
 			if (closure->args[i].a == NULL) {
@@ -1011,7 +1106,7 @@ buffer_size_for_closure(struct wl_closure *closure)
 			}
 
 			size = closure->args[i].a->size;
-			buffer_size += (1 + DIV_ROUNDUP(size, sizeof(uint32_t)));
+			buffer_size += (1 + div_roundup(size, sizeof(uint32_t)));
 			break;
 		default:
 			break;
@@ -1073,11 +1168,11 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 			size = strlen(closure->args[i].s) + 1;
 			*p++ = size;
 
-			if (p + DIV_ROUNDUP(size, sizeof *p) > end)
+			if (p + div_roundup(size, sizeof *p) > end)
 				goto overflow;
 
 			memcpy(p, closure->args[i].s, size);
-			p += DIV_ROUNDUP(size, sizeof *p);
+			p += div_roundup(size, sizeof *p);
 			break;
 		case 'a':
 			if (closure->args[i].a == NULL) {
@@ -1088,11 +1183,11 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 			size = closure->args[i].a->size;
 			*p++ = size;
 
-			if (p + DIV_ROUNDUP(size, sizeof *p) > end)
+			if (p + div_roundup(size, sizeof *p) > end)
 				goto overflow;
 
 			memcpy(p, closure->args[i].a->data, size);
-			p += DIV_ROUNDUP(size, sizeof *p);
+			p += div_roundup(size, sizeof *p);
 			break;
 		default:
 			break;
@@ -1168,19 +1263,22 @@ wl_closure_queue(struct wl_closure *closure, struct wl_connection *connection)
 }
 
 void
-wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
+wl_closure_print(struct wl_closure *closure, struct wl_object *target,
+		 int send, int discarded, uint32_t (*n_parse)(union wl_argument *arg))
 {
 	int i;
 	struct argument_details arg;
 	const char *signature = closure->message->signature;
 	struct timespec tp;
 	unsigned int time;
+	uint32_t nval;
 
 	clock_gettime(CLOCK_REALTIME, &tp);
 	time = (tp.tv_sec * 1000000L) + (tp.tv_nsec / 1000);
 
-	fprintf(stderr, "[%10.3f] %s%s@%u.%s(",
-		time / 1000.0,
+	fprintf(stderr, "[%7u.%03u] %s%s%s@%u.%s(",
+		time / 1000, time % 1000,
+		discarded ? "discarded " : "",
 		send ? " -> " : "",
 		target->interface->name, target->id,
 		closure->message->name);
@@ -1198,11 +1296,23 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 			fprintf(stderr, "%d", closure->args[i].i);
 			break;
 		case 'f':
-			fprintf(stderr, "%f",
-				wl_fixed_to_double(closure->args[i].f));
+			/* The magic number 390625 is 1e8 / 256 */
+			if (closure->args[i].f >= 0) {
+				fprintf(stderr, "%d.%08d",
+					closure->args[i].f / 256,
+					390625 * (closure->args[i].f % 256));
+			} else {
+
+				fprintf(stderr, "-%d.%08d",
+					closure->args[i].f / -256,
+					-390625 * (closure->args[i].f % 256));
+			}
 			break;
 		case 's':
-			fprintf(stderr, "\"%s\"", closure->args[i].s);
+			if (closure->args[i].s)
+				fprintf(stderr, "\"%s\"", closure->args[i].s);
+			else
+				fprintf(stderr, "nil");
 			break;
 		case 'o':
 			if (closure->args[i].o)
@@ -1213,17 +1323,22 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 				fprintf(stderr, "nil");
 			break;
 		case 'n':
+			if (n_parse)
+				nval = n_parse(&closure->args[i]);
+			else
+				nval = closure->args[i].n;
+
 			fprintf(stderr, "new id %s@",
 				(closure->message->types[i]) ?
 				 closure->message->types[i]->name :
 				  "[unknown]");
-			if (closure->args[i].n != 0)
-				fprintf(stderr, "%u", closure->args[i].n);
+			if (nval != 0)
+				fprintf(stderr, "%u", nval);
 			else
 				fprintf(stderr, "nil");
 			break;
 		case 'a':
-			fprintf(stderr, "array");
+			fprintf(stderr, "array[%zu]", closure->args[i].a->size);
 			break;
 		case 'h':
 			fprintf(stderr, "fd %d", closure->args[i].h);
@@ -1234,8 +1349,29 @@ wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send)
 	fprintf(stderr, ")\n");
 }
 
+static int
+wl_closure_close_fds(struct wl_closure *closure)
+{
+	int i;
+	struct argument_details arg;
+	const char *signature = closure->message->signature;
+
+	for (i = 0; i < closure->count; i++) {
+		signature = get_next_argument(signature, &arg);
+		if (arg.type == 'h' && closure->args[i].h != -1)
+			close(closure->args[i].h);
+	}
+
+	return 0;
+}
+
 void
 wl_closure_destroy(struct wl_closure *closure)
 {
+	/* wl_closure_destroy has free() semantics */
+	if (!closure)
+		return;
+
+	wl_closure_close_fds(closure);
 	free(closure);
 }
